/* $Id: x86.c,v 1.9 2004/07/17 00:40:41 mikpe Exp $
 * x86-specific perfctr library procedures.
 *
 * Copyright (C) 1999-2004  Mikael Pettersson
 */
#include <errno.h>
#include <asm/unistd.h>
#include <stdio.h>
#include "libperfctr.h"
#include "x86.h"
#include "x86_cpuinfo.h"

#ifndef __NR_perfctr_info
#ifdef __x86_64__
#define __NR_perfctr_info	247
#else
#define __NR_perfctr_info	284
#endif
#define __NR_vperfctr_open	(__NR_perfctr_info+1)
#define __NR_vperfctr_control	(__NR_perfctr_info+2)
#define __NR_vperfctr_unlink	(__NR_perfctr_info+3)
#define __NR_vperfctr_iresume	(__NR_perfctr_info+4)
#define __NR_vperfctr_read	(__NR_perfctr_info+5)
#endif

#include <unistd.h>

int _sys_perfctr_info(int fd_unused, struct perfctr_info *info, struct perfctr_cpu_mask *cpus, struct perfctr_cpu_mask *forbidden)
{
    return syscall(__NR_perfctr_info, info, cpus, forbidden);
}

int _sys_vperfctr_open(int fd_unused, int tid, int creat)
{
    return syscall(__NR_vperfctr_open, tid, creat);
}

int _sys_vperfctr_control(int fd, const struct vperfctr_control *arg)
{
    return syscall(__NR_vperfctr_control, fd, arg);
}

int _sys_vperfctr_unlink(int fd)
{
    return syscall(__NR_vperfctr_unlink, fd);
}

int _sys_vperfctr_iresume(int fd)
{
    return syscall(__NR_vperfctr_iresume, fd);
}

int _sys_vperfctr_read(int fd, struct perfctr_sum_ctrs *arg1, struct vperfctr_control *arg2, struct perfctr_sum_ctrs *arg3)
{
    return syscall(__NR_vperfctr_read, fd, arg1, arg2, arg3);
}

static int intel_init(const struct cpuinfo *cpuinfo,
		      struct perfctr_info *info)
{
    unsigned int family, model, stepping;

    if (!cpu_has(cpuinfo, X86_FEATURE_TSC))
	return -1;
    family = cpu_family(cpuinfo);
    model = cpu_model(cpuinfo);
    stepping = cpu_stepping(cpuinfo);
    switch (family) {
      case 5:
	if (cpu_has(cpuinfo, X86_FEATURE_MMX)) {
	    /* Avoid Pentium Erratum 74. */
	    if (model == 4 &&
		(stepping == 4 ||
		 (stepping == 3 &&
		  cpu_type(cpuinfo) == 1)))
		info->cpu_features &= ~PERFCTR_FEATURE_RDPMC;
	    return PERFCTR_X86_INTEL_P5MMX;
	} else {
	    info->cpu_features &= ~PERFCTR_FEATURE_RDPMC;
	    return PERFCTR_X86_INTEL_P5;
	}
      case 6:
	if (model == 9 || model == 13)
	    return PERFCTR_X86_INTEL_PENTM;
	else if (model >= 7)
	    return PERFCTR_X86_INTEL_PIII;
	else if (model >= 3)
	    return PERFCTR_X86_INTEL_PII;
	else {
	    /* Avoid Pentium Pro Erratum 26. */
	    if (stepping < 9)
		info->cpu_features &= ~PERFCTR_FEATURE_RDPMC;
	    return PERFCTR_X86_INTEL_P6;
	}
      case 15:
	if (model >= 3)
	    return PERFCTR_X86_INTEL_P4M3;
	else if (model >= 2)
	    return PERFCTR_X86_INTEL_P4M2;
	else
	    return PERFCTR_X86_INTEL_P4;
    }
    return -1;
}

static int amd_init(const struct cpuinfo *cpuinfo,
		    struct perfctr_info *info)
{
    unsigned int family, model, stepping;

    if (!cpu_has(cpuinfo, X86_FEATURE_TSC))
	return -1;
    family = cpu_family(cpuinfo);
    model = cpu_model(cpuinfo);
    stepping = cpu_stepping(cpuinfo);
    switch (family) {
      case 15:
	if (model > 5 || (model >= 4 && stepping >= 8))
	    return PERFCTR_X86_AMD_K8C;
	else
	    return PERFCTR_X86_AMD_K8;
      case 6:
	return PERFCTR_X86_AMD_K7;
    }
    return -1;
}

static int cyrix_init(const struct cpuinfo *cpuinfo,
		      struct perfctr_info *info)
{
    if (!cpu_has(cpuinfo, X86_FEATURE_TSC))
	return -1;
    switch (cpu_family(cpuinfo)) {
      case 6: /* 6x86MX, MII, or III */
	return PERFCTR_X86_CYRIX_MII;
    }
    return -1;
}

static int centaur_init(const struct cpuinfo *cpuinfo,
			struct perfctr_info *info)
{
    unsigned int family, model;

    family = cpu_family(cpuinfo);
    model = cpu_model(cpuinfo);
    switch (family) {
      case 5:
	if (cpu_has(cpuinfo, X86_FEATURE_TSC))
	    return -1;
	info->cpu_features &= ~PERFCTR_FEATURE_RDTSC;
	switch (model) {
	  case 4:
	    return PERFCTR_X86_WINCHIP_C6;
	  case 8: /* WinChip 2, 2A, or 2B */
	  case 9: /* WinChip 3 */
	    return PERFCTR_X86_WINCHIP_2;
	  default:
	    return -1;
	}
      case 6:
	if (!cpu_has(cpuinfo, X86_FEATURE_TSC))
	    return -1;
	switch (model) {
	  case 6: /* Cyrix III */
	  case 7: /* Samuel 2 */
	  case 8: /* Ezra-T */
	  case 9: /* Antaur/Nehemiah */
	    return PERFCTR_X86_VIA_C3;
	  default:
	    return -1;
	}
    }
    return -1;
}

static int generic_init(const struct cpuinfo *cpuinfo,
			struct perfctr_info *info)
{
    if (!cpu_has(cpuinfo, X86_FEATURE_TSC))
	return -1;
    info->cpu_features &= ~PERFCTR_FEATURE_RDPMC;
    return PERFCTR_X86_GENERIC;
}

void perfctr_info_cpu_init(struct perfctr_info *info)
{
    struct cpuinfo cpuinfo;
    int cpu_type;

    identify_cpu(&cpuinfo);
    cpu_type = -1; /* binary compat prevents using 0 for "unknown" */
    if (cpu_has(&cpuinfo, X86_FEATURE_MSR)) {
	switch (cpuinfo.vendor) {
	  case X86_VENDOR_INTEL:
	    cpu_type = intel_init(&cpuinfo, info);
	    break;
	  case X86_VENDOR_AMD:
	    cpu_type = amd_init(&cpuinfo, info);
	    break;
	  case X86_VENDOR_CYRIX:
	    cpu_type = cyrix_init(&cpuinfo, info);
	    break;
	  case X86_VENDOR_CENTAUR:	
	    cpu_type = centaur_init(&cpuinfo, info);
	    break;
	}
    }
    if (cpu_type < 0)
	cpu_type = generic_init(&cpuinfo, info);
    info->cpu_type = cpu_type;
}

unsigned int perfctr_info_nrctrs(const struct perfctr_info *info)
{
    switch( info->cpu_type ) {
#if defined(__x86_64__)
      case PERFCTR_X86_AMD_K8:
      case PERFCTR_X86_AMD_K8C:
	return 4;
      default:
	return 0;
#else
      case PERFCTR_X86_GENERIC:
	return 0;
      case PERFCTR_X86_VIA_C3:
	return 1;
      case PERFCTR_X86_AMD_K7:
      case PERFCTR_X86_AMD_K8:
      case PERFCTR_X86_AMD_K8C:
	return 4;
      case PERFCTR_X86_INTEL_P4:
      case PERFCTR_X86_INTEL_P4M2:
      case PERFCTR_X86_INTEL_P4M3:
	return 18;
      default:
	return 2;
#endif	    
    }
}

const char *perfctr_info_cpu_name(const struct perfctr_info *info)
{
    switch( info->cpu_type ) {
      case PERFCTR_X86_GENERIC:
	return "Generic x86 with TSC";
#if !defined(__x86_64__)
      case PERFCTR_X86_INTEL_P5:
        return "Intel Pentium";
      case PERFCTR_X86_INTEL_P5MMX:
        return "Intel Pentium MMX";
      case PERFCTR_X86_INTEL_P6:
        return "Intel Pentium Pro";
      case PERFCTR_X86_INTEL_PII:
        return "Intel Pentium II";
      case PERFCTR_X86_INTEL_PIII:
        return "Intel Pentium III";
      case PERFCTR_X86_CYRIX_MII:
        return "Cyrix 6x86MX/MII/III";
      case PERFCTR_X86_WINCHIP_C6:
	return "WinChip C6";
      case PERFCTR_X86_WINCHIP_2:
	return "WinChip 2/3";
      case PERFCTR_X86_AMD_K7:
	return "AMD K7";
      case PERFCTR_X86_VIA_C3:
	return "VIA C3";
      case PERFCTR_X86_INTEL_P4:
	return "Intel Pentium 4";
      case PERFCTR_X86_INTEL_P4M2:
	return "Intel Pentium 4 Model 2";
      case PERFCTR_X86_INTEL_PENTM:
	return "Intel Pentium M";
      case PERFCTR_X86_INTEL_P4M3:
	return "Intel Pentium 4 Model 3";
#endif
      case PERFCTR_X86_AMD_K8:
	return "AMD K8";
      case PERFCTR_X86_AMD_K8C:
	return "AMD K8 Revision C";
      default:
        return "?";
    }
}

void perfctr_cpu_control_print(const struct perfctr_cpu_control *control)
{
    unsigned int i, nractrs, nrictrs, nrctrs;

    nractrs = control->nractrs;
    nrictrs = control->nrictrs;
    nrctrs = control->nractrs + nrictrs;

    printf("tsc_on\t\t\t%u\n", control->tsc_on);
    printf("nractrs\t\t\t%u\n", nractrs);
    if( nrictrs )
	printf("nrictrs\t\t\t%u\n", nrictrs);
    for(i = 0; i < nrctrs; ++i) {
        if( control->pmc_map[i] >= 18 ) /* for P4 'fast rdpmc' cases */
            printf("pmc_map[%u]\t\t0x%08X\n", i, control->pmc_map[i]);
        else
            printf("pmc_map[%u]\t\t%u\n", i, control->pmc_map[i]);
        printf("evntsel[%u]\t\t0x%08X\n", i, control->evntsel[i]);
#if !defined(__x86_64__)
        if( control->p4.escr[i] )
            printf("escr[%u]\t\t\t0x%08X\n", i, control->p4.escr[i]);
#endif
	if( i >= nractrs )
	    printf("ireset[%u]\t\t%d\n", i, control->ireset[i]);
    }
#if !defined(__x86_64__)
    if( control->p4.pebs_enable )
	printf("pebs_enable\t\t0x%08X\n", control->p4.pebs_enable);
    if( control->p4.pebs_matrix_vert )
	printf("pebs_matrix_vert\t0x%08X\n", control->p4.pebs_matrix_vert);
#endif
}
